#include "initguid.h"
#include "windows.h"
#include "gdiplus.h"
#include "serial.h"
#include "utilfunc.h"
#include "resource.h"


//{557cf406-1a04-11d3-9a73-0000f81ef32e}
DEFINE_GUID(pngClsid,
	0x557cf406, 0x1a04, 0x11d3, 0x9a, 0x73, 0x00, 0x00, 0xf8, 0x1e, 0xf3, 0x2e);

Gdiplus::Bitmap*    g_Bmp;


Gdiplus::Bitmap* GetBitmap()
{
	return g_Bmp;
}


BOOL GetPixels(HWND hwnd)
{
	TCHAR buf[80];
	BYTE packet_header[4];
	BYTE* packet_body;
	UINT packet_size;
	BYTE more_packets;
	BYTE byte1hi;
	UINT i, j;
	UINT x, y, w, h, screen_w, screen_h, display;
	UINT bits_per_pixel;
	UINT bytes_per_pixel;
	BYTE bit_offset;
	UINT nbytes;
	UINT pixel_repeat;
	UINT color_bits;
	UINT p;
	BOOL bHeader = TRUE;
	BOOL bBppWarningShown = FALSE;
	BYTE R, G, B;
	Gdiplus::Color color;
	HWND hBmp = GetDlgItem(hwnd, IDC_BITMAP);

#ifdef _DEBUG
	HANDLE hDump;
	DWORD dw;
	hDump = CreateFile(TEXT("dump.bin"), GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
#endif

	do {
		if (serial_read((char*)packet_header, sizeof(packet_header)) == -1) {
			MessageBox(hwnd, TEXT("Error reading data"), TEXT("Error"), MB_OK | MB_ICONERROR);
#ifdef _DEBUG
			CloseHandle(hDump);
#endif
			return FALSE;
		}
#ifdef _DEBUG
		WriteFile(hDump, (LPVOID)packet_header, sizeof(packet_header), &dw, NULL);
#endif
		more_packets = packet_header[1] & 0x0f;
		byte1hi = packet_header[1] >> 4 & 0x0f;
		if (packet_header[0] != 0x0a) {
			wsprintf(buf, TEXT("Invalid packet type: %d"), packet_header[0]);
			Log(buf);
			MessageBox(hwnd, buf, TEXT("Error"), MB_OK | MB_ICONERROR);
#ifdef _DEBUG
			CloseHandle(hDump);
#endif
			return FALSE;
		}
		packet_size = (int)(packet_header[2] << 8 | packet_header[3]) - sizeof(packet_header);
		packet_body = (BYTE*)LocalAlloc(NULL, packet_size);
		if (serial_read((char*)packet_body, packet_size) == -1) {
			MessageBox(hwnd, TEXT("Error reading data"), TEXT("Error"), MB_OK | MB_ICONERROR);
			return FALSE;
#ifdef _DEBUG
			CloseHandle(hDump);
#endif
		}

#ifdef _DEBUG
		WriteFile(hDump, (LPVOID)packet_body, packet_size, &dw, NULL);
#endif

		if (bHeader) {
			if (byte1hi == 0) {
				x = packet_body[0];
				y = packet_body[1];
				w = packet_body[2];
				h = packet_body[3];
				screen_w = packet_body[4];
				screen_h = packet_body[5];
				display = packet_body[6];
				bits_per_pixel = packet_body[7];
			} else {
				x = (UINT)packet_body[0] << 8 | packet_body[1];
				y = (UINT)packet_body[2] << 8 | packet_body[3];
				w = (UINT)packet_body[4] << 8 | packet_body[5];
				h = (UINT)packet_body[6] << 8 | packet_body[7];
				screen_w = (UINT)packet_body[8] << 8 | packet_body[9];
				screen_h = (UINT)packet_body[10] << 8 | packet_body[11];
				display = packet_body[12];
				bits_per_pixel = ((UINT)packet_body[13] << 24) | ((UINT)packet_body[14] << 16) | ((UINT)packet_body[15] << 8) | packet_body[16];
			}

			wsprintf(buf, TEXT("%d x %d"), w, h);
			SendDlgItemMessage(hwnd, IDC_LABEL1, WM_SETTEXT, 0, (LPARAM)buf);
			wsprintf(buf, TEXT("%d bpp"), bits_per_pixel);
			SendDlgItemMessage(hwnd, IDC_LABEL2, WM_SETTEXT, 0, (LPARAM)buf);

			bHeader = FALSE;

			bytes_per_pixel = (bits_per_pixel + 7) / 8;
			bit_offset = (BYTE)(bits_per_pixel % 8);
			nbytes = bit_offset ? bytes_per_pixel : bytes_per_pixel + 1;

			if (g_Bmp) delete g_Bmp;
			g_Bmp = new Gdiplus::Bitmap(w, h, PixelFormat24bppRGB);

			int bx, by, ww, wh, dx, dy;
			dx = GetSystemMetrics(SM_CXDLGFRAME) * 2;
			dy = GetSystemMetrics(SM_CYDLGFRAME) * 2 + GetSystemMetrics(SM_CYCAPTION);
			bx = w < 176 ? 101 + (176 - w) / 2 : 101;
			by = h < 220 ? 47 + (220 - h) / 2 : 47;
			ww = w < 176 ? 291 : 291 - 176 + w;
			wh = h < 220 ? 283 : 283 - 220 + h;
			SetWindowPos(hBmp, NULL, bx, by, w+2, h+2, SWP_NOZORDER);
			SetWindowPos(hwnd, NULL, NULL, NULL, ww+dx, wh+dy, SWP_NOZORDER | SWP_NOMOVE);

			p = 0;

		} else {

			for (i = 0; i < packet_size; i += nbytes) {
				color_bits = 0;
				R = G = B = 0;
				for (j = 0; j < bytes_per_pixel; j++) {
					color_bits = packet_body[i + j] | color_bits << 8;
				}
				switch (bits_per_pixel) {
				case 0:
					break;
				case 1:
					R = G = B = color_bits & 0x01 ? 0x00 : 0xFF;
					break;
				case 12:
					R = ((color_bits & 0x0F00) >> 8) * 0x0010;
					G = ((color_bits & 0x00F0) >> 4) * 0x0010;
					B = (color_bits & 0x000F) * 0x0010;
					break;
				case 16:
					R = (color_bits & 0xF800) >> 8 | (color_bits & 0xE000) >> 13;
					G = (color_bits & 0x07E0) >> 3 | (color_bits & 0x0600) >> 9;
					B = (color_bits & 0x001F) << 3 | (color_bits & 0x001C) >> 2;
					break;
				default:
					if (!bBppWarningShown) {
						wsprintf(buf, TEXT("Unsupported color depth: %d bpp\nPlease contact the author"), bits_per_pixel);
						MessageBox(hwnd, buf, TEXT("Warning"), MB_OK | MB_ICONWARNING);
						bBppWarningShown = TRUE;
					}
					break;
				}
				color = Gdiplus::Color(R, G, B);
				pixel_repeat = bit_offset ? (UINT)packet_body[i] >> bit_offset : (UINT)packet_body[i + bytes_per_pixel];
				for (j = 0; j <= pixel_repeat; j++) {
					g_Bmp->SetPixel(p % w, p / w, color);
					p++;
				}
			}
			InvalidateRect(hBmp, NULL, NULL);
		}
		LocalFree(packet_body);

	} while (more_packets);

#ifdef _DEBUG
	CloseHandle(hDump);
#endif

	return TRUE;
}


void SaveImage(LPTSTR lpBaseName)
{
	Gdiplus::Status stat;

	TCHAR fname[MAX_PATH];
	int cnt = 0;
	HANDLE hFile;
	if (!lpBaseName[0]) {
		lstrcpy(lpBaseName, TEXT("screen"));
	}
	while(1) {
		wsprintf(fname, TEXT("%s%d.png"), lpBaseName, ++cnt);
		hFile = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE)
			break;
		else
			CloseHandle(hFile);
	}
	Log(TEXT("Saving image"));
	Log(fname);
#ifdef UNICODE
	stat = g_Bmp->Save(fname, &pngClsid, NULL);
#else
	WCHAR fnameW[MAX_PATH];
	MultiByteToWideChar(CP_ACP, NULL, fname, -1, fnameW, MAX_PATH);
	stat = g_Bmp->Save(fnameW, &pngClsid, NULL);
#endif
	if (stat != Gdiplus::Ok) {
		Log(TEXT("Error saving image"));
		MessageBox(NULL, TEXT("Error saving image"), TEXT("Error"), MB_ICONERROR | MB_OK | MB_TASKMODAL);
	}

#ifdef _DEBUG
	wsprintf(fname, TEXT("%s%d.bin"), lpBaseName, cnt);
	Log(TEXT("renaming dump.bin"));
	Log(fname);
	if (!MoveFile(TEXT("dump.bin"), fname)) {
		Log(TEXT("Error renaming dump.bin"));
	}
#endif
}
