#include "windows.h"
#include "utilfunc.h"


HANDLE hComm;


// Open the serial port and store the settings.
int serial_open(LPCTSTR file)
{
	DCB dcb;

	if ((hComm = CreateFile(
			file,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_NO_BUFFERING,
			NULL)) == INVALID_HANDLE_VALUE) {
		return FALSE;
	} else {

		// get any early notifications
//		if (!SetCommMask(hComm, EV_RXCHAR))
//			ErrorExit(TEXT("SetCommMask"));

		// setup device buffers
		if (!SetupComm(hComm, 4096, 4096))
			ErrorExit(TEXT("SetupComm"));

		// purge any information in the buffer
		if (!PurgeComm(hComm, PURGE_TXABORT | PURGE_RXABORT |
			  PURGE_TXCLEAR | PURGE_RXCLEAR))
			ErrorExit(TEXT("PurgeComm"));

/*		COMMTIMEOUTS timeouts;

		timeouts.ReadIntervalTimeout = MAXDWORD;
		timeouts.ReadTotalTimeoutMultiplier = 0;
		timeouts.ReadTotalTimeoutConstant = 0;
		timeouts.WriteTotalTimeoutMultiplier = 0;
		timeouts.WriteTotalTimeoutConstant = 0;

		if (!SetCommTimeouts(hComm, &timeouts))
			ErrorExit(TEXT("SetCommTimeouts"));
*/
		if (!GetCommState(hComm, &dcb))
			ErrorExit(TEXT("GetCommState"));

		// Fill in DCB: 256,000 bps, 8 data bits, no parity, and 1 stop bit.
		dcb.BaudRate = CBR_256000;		// set the baud rate
		dcb.ByteSize = 8;						 // data size, xmit, and rcv
		dcb.Parity = NOPARITY;				// no parity bit
		dcb.StopBits = ONESTOPBIT;		// one stop bit

		if (!SetCommState(hComm, &dcb))
			ErrorExit(TEXT("SetCommState"));
	}

	return TRUE;
}


// Close the serial port and restore old settings.
int serial_close()
{
	// disable event notification and wait for thread
	// to halt
	//SetCommMask(hComm, 0);

	// drop DTR
	EscapeCommFunction(hComm, CLRDTR);

	// purge any outstanding reads/writes and close device handle
	PurgeComm(hComm, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);

	CloseHandle(hComm);

	return TRUE;
}


// Read from serial device.
int serial_read(char *lpBuf, int nRead)
{
	DWORD dwRead;
	DWORD l = 0;
	DWORD ticks = GetTickCount();

	int nSleeps = 0;
	while(1) {
		if (!ReadFile(hComm, (LPVOID)(lpBuf + l), (nRead - l), &dwRead, NULL)) {
			Log(TEXT("serial_read(): error"));
			return -1;
		}
		l += dwRead;
		if (dwRead) {
			ticks = GetTickCount();
		}
		if (l < nRead) {
			if (GetTickCount() - ticks >= 5000) {
				TCHAR buf[1024];
				wsprintf(buf, TEXT("serial_read(): timeout; %d bytes expected, %d bytes read"), nRead, l);
				Log(buf);
				break;
			}
			nSleeps++;
			Sleep(1);
		} else {
			break;
		}
	}
	if (nSleeps) {
		TCHAR buf[1024];
		wsprintf(buf, TEXT("serial_read(): %d wait cycles, %d bytes read"), nSleeps, nRead);
		Log(buf);
	}
	return l;
}


// Write to serial device.
int serial_write(char *lpBuf, int nWrite)
{
	DWORD dwWritten;
	DWORD l = 0;
	DWORD ticks = GetTickCount();

	int nSleeps = 0;
	while(1) {
		if (!WriteFile(hComm, (LPVOID)(lpBuf + l), (nWrite - l), &dwWritten, NULL)) {
			Log(TEXT("serial_write(): error"));
			return -1;
		}
		l += dwWritten;
		if (dwWritten) {
			ticks = GetTickCount();
		}
		if (l < nWrite) {
			if (GetTickCount() - ticks >= 5000) {
				TCHAR buf[1024];
				wsprintf(buf, TEXT("serial_write(): timeout; %d bytes queued, %d bytes written"), nWrite, l);
				Log(buf);
				break;
			}
			nSleeps++;
			Sleep(1);
		} else {
			break;
		}
	}
	if (nSleeps) {
		TCHAR buf[1024];
		wsprintf(buf, TEXT("serial_write(): %d wait cycles, %d bytes written"), nSleeps, nWrite);
		Log(buf);
	}
	return l;
}

BOOL send_at_command(char *cmd)
{
	int l = lstrlenA(cmd) + 5;
	char *buf = (char *)LocalAlloc(NULL, l);
	wsprintfA(buf, "AT%s\r\n", cmd);
	TCHAR t[1024];
	wsprintf(t, TEXT("AT%hs"), cmd);
	Log(t);
	BOOL ret = TRUE;
	if (serial_write(buf, l-1) == -1) {
		ret = FALSE;
	}
	LocalFree(buf);
	return ret;
}	

BOOL get_modem_reply(char *cmd)
{
	DWORD nread;
	int ld;
	char *reply;
	char *buffer;
	COMSTAT ComStat;
	DWORD dwErrorFlags;
	DWORD ticks = GetTickCount();

	nread = 0;
	reply = NULL;
	buffer = NULL;
	ld = 0;

	int nSleeps = 0;
	// send command to modem
	if (cmd) {
		if (!send_at_command(cmd)) {
			return FALSE;
		}
	}

	// wait for data to arrive on the line
	while (1) {
		while (1) {
			// how much data do we have?
			ClearCommError(hComm, &dwErrorFlags, &ComStat);
			nread = ComStat.cbInQue;
			if (nread) break;
			if (GetTickCount() - ticks >= 5000) {
				Log(TEXT("get_modem_reply(): timeout"));
				LocalFree(buffer);
				LocalFree(reply);
				return FALSE;
			}
			nSleeps++;
			Sleep(10);
		}

		// we know what the data size is - alloc space for it
		buffer = (char*)LocalAlloc(LPTR, nread);
		if (!buffer)
			ErrorExit(TEXT("LocalAlloc"));

		if (serial_read(buffer, nread) == -1) {
			LocalFree(buffer);
			LocalFree(reply);
			return FALSE;
		}
		ticks = GetTickCount();

		// append it to reply
		ld = bufappend(&reply, ld, buffer, nread);

		LocalFree(buffer);

		// search for msg
		char msg_error[] = "\r\nERROR\r\n";
		char msg_ok[]    = "\r\nOK\r\n";
		if (memmem(reply, ld, msg_error, sizeof(msg_error)-1)) {
			if (nSleeps) {
				TCHAR buf[1024];
				wsprintf(buf, TEXT("get_modem_reply(): %d wait cycles"), nSleeps);
				Log(buf);
			}
			Log(TEXT("ERROR"));
			LocalFree(reply);
			return FALSE;
		}
		else if (memmem(reply, ld, msg_ok, sizeof(msg_ok)-1)) {
			if (nSleeps) {
				TCHAR buf[1024];
				wsprintf(buf, TEXT("get_modem_reply(): %d wait cycles"), nSleeps);
				Log(buf);
			}
			Log(TEXT("OK"));
			LocalFree(reply);
			return TRUE;
		}
	}
}
